//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 990422, vfb, original Java
 * Translated from the Swarm Objective-C Gecko source, version 0.51
 */
/** A Metabolism is an inventory of resources, plus the logic to
 *  metabolize, which is turning needed input resources into body mass.
 * Discussion:  an agent must convert resources from its gullet to its
 * stores in order to assimilate anything it's eaten/gotten.  If it's 1:1,
 * that's a conversion from A, to A, rate 1.0.  In general, anything
 * can be converted to anything, and this supports a conversion matrix.  But
 * almost always, the conversion matrix isn't needed, so we cheap out whenever
 * possible.  An agent's size is lock-step to its stores, but that logic must
 * go uplevel to Critter, as we don't have a generic way to convert stores of
 * an arbitrary number of resources to one radius.
 */


package gecko;

public class Metabolism {
  private ResourceList gullet;
  private ResourceList stores;
  private double[][]  convert;
  private double[]   taxrate;		// taxes are of form taxrate * biomass ^ taxexp
  private double[]   taxexp;

  // these are set to nullres if the story is complicated
  private int ieat;	  	  // what I eat
  private int iam;	          // what I convert it to
  private double efficiency;	  // if ieat and iam, conversion is simple, too.

  private static int numRes = Globals.getResources().getNumResources();  // so long as it doesn't change...
  private static int nullRes = 0;  // so long as it doesn't change...

  public Metabolism() {
    init();
  }

  /** Sets all fields to zero. */
  public void zero() {
    int i, j;
    gullet.zero();
    stores.zero();
    if (convert != null) {
      for (i=0; i<numRes; i++)
        for (j=0; j<numRes; j++)
          convert[i][j] = 0.0;
    }
    for (i=0; i<numRes; i++) {
      taxrate[i] = 0.0;
      taxexp[i]  = 0.0;
    }
    ieat 	     = nullRes;
    iam 	     = nullRes;
    efficiency = 0.0;
  }

  protected void init() {
    gullet  = new ResourceList();
    stores  = new ResourceList();
    taxrate = new double[numRes];
    taxexp  = new double[numRes];

    // don't allocate convert matrix until we need one.
    convert = null;
    zero();
  }

  /** Add a conversion factor. If more than one, the conversion matrix starts. */
  public void addConversion(int fromRes, int toRes, double rate) {
    int i, j;
      //
     // Is this the first (only) conversion?
    // 
    if ( (ieat==nullRes) && (convert==null) ){
      ieat       = fromRes;
      iam        = toRes;
      efficiency = rate;
      //
     // This is not the first (only) conversion.
    // 
    } else {
        //
       // Do we have a convert matrix going yet?
      //
      if (convert==null) {
          //
         // No, so allocate one.
        //
        convert = new double[numRes][];
        for (i=0; i<numRes; i++)
          convert[i] = new double[numRes];
        for (i=0; i<numRes; i++)
          for (j=0; j<numRes; j++)
            convert[i][j] = 0.0;
          //
         // And transfer prior simple conversion into convert matrix.
        //
        convert[ieat][iam] = efficiency;
        ieat = iam = nullRes;
        efficiency = 0.0;
      }
        //
       // Add the current conversion into convert matrix.
      //
      convert[fromRes][toRes] = rate;
    }
  }

  /** Add or replace this conversion.*/
  public void setConversion(int fromRes, int toRes, double newrate) {
    // Is this the first (only) conversion?
    if ( (ieat==nullRes) && (convert==null) ) {
      addConversion(fromRes, toRes, newrate);
      return;

    // Not the first (only) conversion.
    } else {
      // No convert matrix going yet?
      if (convert==null) {
        // Replace sole conversion?
        if ( (ieat == fromRes) && (iam == toRes) ) {
          efficiency = newrate;
          return;

        // No, second conversion needed--use add routine.
        } else {
          addConversion(fromRes, toRes, newrate);
          return;
        }

      // Yes, have matrix - simply stuff this in.
      } else {
        convert[fromRes][toRes] = newrate;
      }
    }
  }

  public void setTax(int resource, double rate, double exponent){
    taxrate[resource] = rate;
    taxexp[resource]  = exponent;
  }

  public double getTaxRate(int resource) {return taxrate[resource];}
  public double getTaxExponent(int resource) {return taxexp[resource];}
  public double getConversion(int fromRes, int toRes) {
    if ( (ieat == fromRes) && (iam == toRes) )
      return efficiency;
    else if (ieat != nullRes)
      return 0.0;
    else if (convert != null)
      return convert[fromRes][toRes];
    else
      return 0.0;
  }

  /** never copy gullet--only the rules are transferable. */
  public void setFrom(Metabolism other) {
    int i, j;

    for (i=0; i<numRes; i++) {
      taxrate[i] = other.taxrate[i];
      taxexp[i]  = other.taxexp[i];
    }
    ieat           = other.ieat;
    iam            = other.iam;
    efficiency     = other.efficiency;

    if (other.convert != null) {
      convert = new double[numRes][];
      for (i=0; i<numRes; i++)
        convert[i] = new double[numRes];
      for (i=0; i<numRes; i++)
        for (j=0; j<numRes; j++)
          convert[i][j] = other.convert[i][j];
    }
    stores.setFrom(other.stores);
  }

  /** metabolize gullet -> stores */
  public void metabolize() {
    int i, j;
    
    // simple
    if (ieat>nullRes) {
      stores.list[iam] += gullet.list[ieat] * efficiency;
      gullet.list[ieat] = 0.0;

    // complex
    } else {
      for (i=1; i<numRes; i++) {
        for (j=1; j<numRes; j++) {
          if (convert[i][j] > 0) {
            stores.list[j] += gullet.list[i] * convert[i][j];
          }
        }
      }
      for (i=1; i<numRes; i++) {
        gullet.list[i] = 0.0;
      }
    }
  }

  /** Debit taxes and return list to pay. Ie, they're subtracted here, and
   *  amount list is returned, for adding wherever they're paid -to-.
   */
  public ResourceList payTaxes(){
    int i;
    ResourceList topay = new ResourceList();

    if (iam != nullRes) {
      topay.list[iam] = taxrate[iam] * Math.pow(stores.list[iam], taxexp[iam]);
    } else {
      for (i=1; i < numRes; i++) {
        topay.list[i] = taxrate[i] * Math.pow(stores.list[i], taxexp[i]);
      }
    }

    stores.subtract(topay);
    return topay;
  }
  public double getStores(int resource) { return stores.list[resource]; }

  public ResourceList getStores() { return stores; }

  public void setStores(ResourceList list){stores.setFrom(list);}

  public void setStores(int resource, double quant){stores.list[resource] = quant;}

  public int getIeat() { return ieat; }

  public int getIam()  { return iam; }

  public void setIeat(int what) {ieat=what;}
  public void setIam(int what) {iam=what;}

  public boolean doIeat(int what){
    if (ieat != nullRes) {
      if (ieat == what)
        return true;
    } else {
      if (convert==null)
        return false;
      for (int i=1; i<numRes; i++)
        if (convert[what][i] > 0.0)
          return true;
    }
    return false;
  }

  public boolean amI(int what){
    if (iam != nullRes) {
      if (iam == what)
        return true;
    } else {
      if (convert==null)
        return false;
      for (int i=1; i<numRes; i++)
        if (convert[i][what] > 0.0)
          return true;
    }
    return false;
  }
  public void feed(ResourceList dinner) {gullet.add(dinner);}

  public void feed(int resource, double amount){gullet.list[resource] += amount;}

  public void debit(ResourceList payout){stores.subtract(payout);}

  public String toString() {
    if (iam != nullRes) {
      return "Metabolism[converts " + ieat + " to " + iam + " at " + efficiency + ",stores=" + stores
             + ",gullet=" + gullet + "]";
    } else {
      return "(conversion matrix)";
    }
  }
}

