//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator

/** Collects the aspect of the landscape site best dealt with for the sites
 *  as a whole--statistics, name, etc. Site is analogous to Species.
 *  At time 0, the prototype site is cloned and hung on the quad tree, if it
 *  is a Quad. Its initial mcr should be set in the Quad itself.
 *
 *  Although Site is a StatSource, that's for birth/death listeners.
 *  In practice, Gecko doesn't use those listeners. Though they might be
 *  useful if species could emerge or go extinct in the course of a run.
 *  What they're actually for in the stat package is for notifying things
 *  that refer to the StatSource, that the StatSource is going away or a
 *  new one needs to be made. For current Gecko needs, that doesn't happen,
 *  as Sites are specified and eternal at run begin.
 */

package gecko;

import stats.*;
import java.util.*;
import java.lang.reflect.*;
import java.awt.Color;
import gecko.statglue.*;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.lang.reflect.*;

public class Site extends StatSource implements Biomassable {
  static private Vector allSites = new Vector(1,1);
  private Siteable prototype; // prototype site (copied to make real sites)
  private String name;
  private Siteable instance;  // instance of prototype that was created, if any...
  private int colorId; // set from prototype if possible; else pink.

  // Statistics
  private double population = 0; // if anyone wants population...dunno why...
  private double biomass = 0; // scaled total resources of members
  private double biomassFed = 0;	// total fed out of members

  /** Similarly to Species, must call register to add this to list of
   *  all sites. */
  public Site(String name, Siteable prototype, Color c) {
    super();
    setPrototype(prototype);
    colorId = QuadDisplay.setColor(name, c);
  }

  public void register() {
    if (!allSites.contains(this))
      allSites.addElement(this);
    resetStats();
  }

  static public void removeAllSites(){allSites.removeAllElements();}
  static public Vector getAllSites() { return allSites; }
  static public Site find(String name) {
    Site s;
    for (int i=0; i< allSites.size(); i++) {
      s = (Site) allSites.elementAt(i);
      if (s.getName().equals(name))
        return s;
    }
    return null;
  }

  private ExecParam makeEP(String observeWhat) {
    ExecParam ep;
    try {
      Method observeM = this.getClass().getMethod("observe" + observeWhat,
        Param.NO_ARGS_SIGNATURE);
      ep = new ExecParam(observeWhat, Param.NO_ARGS, this, observeM);
    } catch (Exception e) {
      System.err.println("Couldn't observe " + name + " statistic " + observeWhat);
      return null;
    }
    return ep;
  }

  private ExecParam makeTaxEP(int whichResource, String resourceName, Color c) {
    ExecParam ep;
    try {
      Class[] signature = { String.class, int.class, Color.class };
      Object[] args = { resourceName, new Integer(whichResource), c };
      Method observeM = this.getClass().getMethod("observeBiomassTax", signature);
      ep = new ExecParam(resourceName + " Taxes", args, this, observeM);
    } catch (Exception e) {
      System.err.println("Couldn't observe " + resourceName + " taxes");
      return null;
    }
    return ep;
  }

  /** Build list of ExecParam's to turn on stats for one species. */
  public ParamList getStatParamList() {
    ParamList answer = new ParamList();
    answer.setName("Site Stats");
    answer.add(makeEP("Biomass"));
    answer.add(makeEP("BiomassFed"));
    // this one's harder...
    Resourceable res = Globals.getResources();
    int maxres = res.getNumResources();
    Palette palette = new PaletteEarthTones(maxres);
    for (int i=1; i<maxres; i++) {
      answer.add( makeTaxEP(i, res.getResourceName(i), palette.getColor(i-1) ) );
    }
    return answer;
  }

  static public void resetStatsAll() {
    Site s;
    for (int i=0; i< allSites.size(); i++) {
      s = (Site) allSites.elementAt(i);
      s.resetStats();
    }
  }

  public void resetStatsRound() {
    biomass = biomassFed = 0;
  }

  public void resetStats() {
    population = 0;
    resetStatsRound();
  }

  public void setPrototype(Siteable proto) {
    prototype = proto;
    prototype.setSite(this);
    resetStats();
  }

  /** Step just resets statistics. */
  public void step() {
    resetStatsRound();
  }
  /** Step just resets statistics. */
  static public void stepAll() {
    for (int i=0; i<allSites.size(); i++)
      ((Site)allSites.elementAt(i)).step();
  }

  public String toString() {
    return "Site[name=" + name + ",prototype=" + prototype;
  }

  public String getName() {return name;}
  public int getColorId() {return colorId;}
  public Color getColor() {return QuadDisplay.getColor(colorId);}
  public void setName(String s){name=s;}


  static public void hangAllSites() {
    for (int i=0; i<allSites.size(); i++)
      ((Site)allSites.elementAt(i)).birthday();
  }
  /** On day 0, birthday hangs one newborn copy of prototype on the world tree,
   *  if it is a Quad. Otherwise, just instantiates the prototype, which...
   *  just *exists.* */
  protected void birthday() {
    resetStats();
    try {
      instance = (Siteable) prototype.newCopy();
    } catch (Exception e) {
    }
    if (instance instanceof Quad) {
      Globals.addLandQuad((Quad) instance);
    }
    if ((instance instanceof Mcr) && ((Mcr)prototype).isNullMcr() ) {
      if (Globals.getWorldMcr() != null) {
        ((Mcr) instance).setMcr(Globals.getWorldMcr());
      }
    }
  }

  public double getPopulation() { return population; }
  public void notifyBirth() { population += 1.0; }
  public void notifyDeath() { population -= 1.0; }
  /** At the end of each step, each member of the site notifies the Site
   * what its current biomass/etc is. Biomass/etc for the Site is zeroed by step.
   */
  public void notifyBiomass(double b) { biomass += b; }
  public double getBiomass() { return biomass; }
  public void notifyBiomassFed(double b) { biomassFed += b; }
  public double getBiomassFed() { return biomassFed; }

  private void observeCommon(String plotname, String name, StatSourceBroker broker, Color c) {
    MetaPlot mp = Globals.getStats().getMetaPlot(plotname);
    if (mp.hasTrace(name)) {
      return;
    }
    StatTrace st = new StatTrace(name, broker);
    st.setXTime(new Globals());
    st.setColor(c);
    mp.addTrace(st);
    Globals.getUI().makeStatDumpers(mp);
  }

  public void observeBiomass() {
    BiomassBroker broker = new BiomassBroker(this, prototype.getResource());
    observeCommon(Globals.BIO_PLOT_NAME,
      Globals.getResources().getResourceName(prototype.getResource()) + " Biomass",
      broker, QuadDisplay.getColor(colorId) );
  }
  public void observeBiomassFed() {
    BiomassFedBroker broker = new BiomassFedBroker(this, prototype.getResource());
    observeCommon(Globals.BIO_PLOT_NAME,
      Globals.getResources().getResourceName(prototype.getResource()) + " BiomassFed",
      broker, QuadDisplay.getColor(colorId) );
  }

  static public double getBiomassTax(int resourceId) {
    if (Globals.getLandRoot() == null)
      return 0;
    return Globals.getLandRoot().getTaxesPaid(resourceId);
  }

  public void observeBiomassTax(String label, int resourceId, Color c) {
    BiomassTaxBroker broker =
      new BiomassTaxBroker(this, resourceId);
    QuadDisplay.setColor(label, c);
    observeCommon(Globals.BIO_PLOT_NAME, label, broker, c );
  }
}
