//Title:        BacSim/Gecko
//Version:
//Copyright:    Copyright (c) 1997-1999, All rights reserved
//Author:       Jan Kreft, Kreft@cardiff.ac.uk
//Company:      University of Cardiff
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Bacterial Ecology Simulator (BacSim)
/**
 * BacShove keeps the "doTheLocomotion" logic common to Eps and Bacillus.
 * 990928, vfb, translated from the Swarm Objective-C BacSim source.
 * 001212, vfb, update shove translation
 */


package gecko.bacsim;
import gecko.*;
import gecko.zoo.*;

public class BacShove {

  private static McBrick shovewin  = new McBrick();

  /* apply boundary conditions, biofilm and wrap around */
  static public void wrapXYZ(QuadCircleSpecies critter, Diffuse actualSite) {

    if (actualSite.isBiofilm()) {
      // biofilms grow on carrier but no agent should be within the carrier
      if (critter.x <= actualSite.getCarrierLayer()) {
        critter.x = actualSite.getCarrierLayer() + BacGlobals.EPSILON;
      }
    }

    if (actualSite.isPeriodicX()) {
      double lengthX = actualSite.getLengthX();
      if (critter.x < 0) critter.x += lengthX;
      if (critter.x >= lengthX) critter.x -= lengthX;
    }

    if (actualSite.isPeriodicY()) {
      double lengthY = actualSite.getLengthY();
      if (critter.y < 0) critter.y += lengthY;
      if (critter.y >= lengthY) critter.y -= lengthY;
    }

    if (actualSite.isPeriodicZ()) {
      double lengthZ = actualSite.getLengthZ();
      if (critter.z < 0) critter.z += lengthZ;
      if (critter.z >= lengthZ) critter.z -= lengthZ;
    }

  }


  /** "mover" hunts under myRoot for all neighbors to be shoved by. shoveRadius
   *  is the -absolute- radius of the search window applied by each agent
   *  to find its neighbors. (E.g., for Bacillus it would be shoveRadius*radius,
   *  because Bacillus' shoveRadius is a -multiplier-, and this is the -result-.)
   *  The distance shoved is simply the vector sum of overlap radii.
   */
  public static void shove(QuadCircleSpecies mover, Diffuse actualSite, double shoveRadius)
  {
    Critter neighbor;
    Quad nq;
    double overlapRadius, distRadius;
    Quad mobileRoot   = Globals.getMobileRoot();
    boolean periodicY = actualSite.getPeriodicY();
    double lengthY    = actualSite.getLengthY();
    double lengthZ    = actualSite.getLengthZ();
    Direction wanna   = ((GetDirectionable)mover).getDirection();
    if (!Bacillus.mutualShove)
      wanna.zeroDirection();

    // pick around for neighbors inside distRadius
    mover.setVisible(false);
    if ( !(mover instanceof Bacillus) ) {
      shovewin.sphereToBrick(mover.x, mover.y, mover.z, shoveRadius);
    } else {
      Bacillus bacillus = (Bacillus) mover;
      // initialize shove window
      // if we are producing eps, our distance to neighbors is explicitly
      // modelled by the total volume taken up by
      // EPS and biomass, else we multiply our radius by a calibrated factor
      // to avoid overlap between cells
      switch (bacillus.getEps()) {
        case Eps.NOEPS:
          distRadius = shoveRadius * bacillus.radius;
          shovewin.sphereToBrick(mover.x, mover.y, mover.z, distRadius);
          break;
        case Eps.CAPSULE:
          distRadius = shoveRadius * bacillus.getEpsRadius();
          shovewin.sphereToBrick(mover.x, mover.y, mover.z, distRadius);
          break;
        case Eps.SLIME:
          distRadius = shoveRadius * bacillus.getEpsRadius();
          shovewin.sphereToBrick(mover.x, mover.y, mover.z, distRadius);
          break;
        case Eps.SELFBINDING:
        case Eps.ALLBINDING:
        case Eps.SLIMERSBINDING:
          distRadius = shoveRadius * bacillus.getEpsRadius();
          // need larger shovewin if we also deal with attractive forces
          shovewin.sphereToBrick(mover.x, mover.y, mover.z, 3.0*distRadius);
          break;
      }
    }
    nq = mobileRoot.pickWin(shovewin, mobileRoot, false);
    while (nq != null) {
      // get one shove
      neighbor = (Critter) nq;
      overlapRadius = shoveRadius;
      overlapRadius -= Globals.distance( mover.x, mover.y, mover.z,
          neighbor.x, neighbor.y, neighbor.z )
          - neighbor.radius;
      if (overlapRadius > 0) {
        // notice that only mover is moving--the neighbor isn't taking instruction now
        if (Bacillus.isSquareShove())
          overlapRadius *= overlapRadius;

        wanna.addDirection(mover.x - neighbor.x,
                          mover.y - neighbor.y,
                          mover.z - neighbor.z, overlapRadius);
      }
      // next
      nq = nq.pickWin(shovewin, mobileRoot, false);
    }
    mover.setVisible(true);

    if (periodicY) {
      // Then get a shove from the concrete wall called carrier as well
      // (none if we don't sit on carrier)
      // carrier is always at the low x side
      overlapRadius = actualSite.getOverlapWithCarrierAt((mover.x - shoveRadius), mover.y);
      if (overlapRadius > 0) {
        //[self addDirectionX: 1 Y: 0 Weight: overlapRadius];
        wanna.addDirection(1, 0, 0, overlapRadius);
      }
    }

    // now do it--getWannas returns the current values, no normalization!
    double[] u = wanna.getWannas();
    mover.x += u[0];
    mover.y += u[1];
    mover.z += u[2];

    wrapXYZ(mover, actualSite);

    mover.setMcr();
  }

  /** percolate shoves--Diffuse tells us when. The site is needed because
   *  it provides the "calcPorosity" routine that's consulted for the
   *  end condition.
   */
  public static double doTheLocomotion(Diffuse actualSite) {
    int count = 0;
    Quad q;
    boolean relaxed = false;
    double minPorosity = 0;
    Quad mobileRoot = Globals.getMobileRoot();
    Bacillus bacillus;
    Eps eps;

    while (!relaxed) {
      count++;
      //traverse the agent quadding tree and tell everyone to move, including Eps
      q = mobileRoot.pickWin(mobileRoot, mobileRoot, false);
      while (q != null) {
        if (q instanceof Bacillus) {
          bacillus = (Bacillus) q;
          double distRadius;
          if (bacillus.isEps())
            distRadius = bacillus.getEpsRadius();
          else
            distRadius = bacillus.getShoveRadius() * bacillus.radius;
          shove(bacillus, actualSite, distRadius); // this does the actual work
        } else if (q instanceof Eps) {
          eps = (Eps) q;
          shove( eps, actualSite, eps.getRadius() );
        }

        q = q.pickWin(mobileRoot, mobileRoot, false);
      }
      minPorosity = actualSite.calcPorosity();

      if (minPorosity > 0.26)
        relaxed = true;
      if (count > 1000)
        Globals.log("WARNING: made over 1000 loco-loops, overcrowded?");
    }
    return minPorosity;
  }

}