//Title:        BacSim/Gecko
//Version:
//Copyright:    Copyright (c) 1997-2001, All rights reserved
//Author:       Jan Kreft, Kreft@cardiff.ac.uk
//Company:      University of Cardiff
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Bacterial Ecology Simulator (BacSim)
/**
 * Eps is Extracellular Polymeric Substance, earlier called Extracellular PolySaccharide
 * 990927, vfb, translated from the Swarm Objective-C BacSim source.
 * 001209, vfb, update translation
 */

package gecko.bacsim;

import gecko.zoo.*;
import gecko.*;
import java.io.*;
import stats.*;
import courseware.util.*;

public class Eps extends QuadCircleSpecies implements GetDirectionable, HasProcess {
//public class Eps extends Critter {

  // from objc EpsType
  public final static int NOEPS = 0;
  public final static int CAPSULE = 1;
  public final static int SLIME = 2;
  public final static int SELFBINDING = 3;
  public final static int ALLBINDING = 4;
  public final static int SLIMERSBINDING = 5;

  // personal
  // this is the master variable, radius is set from epsDryMass, and initVol is set from epsDryMass upon creation of instance
  private double epsDryMass;    // du jour
  // initVol is the initial volume of this object, never changes once set, but may be different from instance to instance
  private double initVol;       // vfb, 001227: why not static? Ginger: see above
  private Direction direction = new Direction(); // a feature of Critter, not QuadCircleSpecies

  // global parameters valid for all Eps blobs, set by Protocol!
  private static double epsDensity;     // the density of the EPS (for epsDryMass <-> volume conversions)
  private static double convertBack;		// when decay or die, multiplier on mass->substrate
  private static double decayR;		      // rate of Eps decay

  private static Eps anymember;         // anymember of the species, last in practice.

  // vfb opinionated: quit copying diffuse's stuff all over, the bookkeeping is annoying....
  private static Diffuse actualSite;
  // avoid allocating this thing over and over
  private static McBrick shovewin = new McBrick(); // not needed when shoving done by BacShove???

  /** This property list is used to generate paramLists for popup editing.
   *  Note initArea is left off for now, as don't have the routines to set
   *  it nicely from a popup. Note also that you could have radius OR
   *  epsDryMass on the list, not both. Whichever is most convenient for
   *  specification (they are locked.)
   */
  protected static String[] props = { "epsDensity", "convertBack", "radius", "decayR" };
  public String[] getPropertyNames() { return props;}

  public Direction getDirection() { return direction; }

  public Eps() {
    super();
    anymember = this;
  }

  public boolean putOnPlantRoot() {return false;}

  // setFrom not needed, since epsDryMass isn't transferrable & rest global
  public Object clone() {
    return makeKid();
  }

  public Speciesable makeKid() {
    Eps kid = new Eps();
    kid.setFrom(this);
    //kid.setEpsDryMass(epsDryMass); // ensures radius matches Ginger: Why needed?
    // since using bogus parent, let's not keep tabs!
    // getSpecies().notifyBirthToMom(this);
    return (Speciesable) kid;
  }

  public static void notifyBiomass(double m) {
    anymember.getSpecies().notifyBiomass(m);
  }

  /** Translation: Creates a new Eps of given biomass. */
  public static void excreteEpsAmount(double e, double xi, double yi, double zi) {
    Eps baby;
    // make a kid
    baby = (Eps) anymember.makeKid();

    baby.birthday = Globals.getCurrentTime();

    baby.epsDryMass = e;
    baby.initVol    = e / epsDensity;
    baby.radius = Globals.volume2radius( baby.initVol );

    baby.x      = xi;
    baby.y      = yi;
    baby.z      = zi;

    // out of bounds checking
    BacShove.wrapXYZ(baby,actualSite);
    // hang on quad tree
    baby.setMcr();
    Globals.getMobileRoot().addNewbie(baby);
    baby.getSpecies().notifyBirth();
  }

  /** Eps does this investigation and tells us for bookkeeping convenience. Great
   *  place for any and all run initialization stuff.
   */
  public static void setSiteInfo(Diffuse site) {
    actualSite = site;
  }


  public void setEpsDensity(double v) {epsDensity = v;}

  public double getEpsDensity() {return epsDensity;}

  public void setConvertBack(double v) {convertBack = v;}

  public double getConvertBack() {return convertBack;}


  /** translation:  lock radius and epsDryMass, so that setting either is

   *  equivalent to setting both.

   *  The results of these depend on epsDensity being set first.

   *  This boils down to my translation of a bunch of things--kids are

   *  born at size radius of progenitor, now, not kidsize, when Species

   *  instantiates them into the world.

   */

  public void setRadius(double v) {

    radius = v;

    epsDryMass = epsDensity * Globals.radius2volume(v);

  }

  public double getRadius() {return radius;}


  // translation, vfb, 001227: came from setX:Y:R in Eps.m

  public void setEpsDryMass(double v) {

    epsDryMass = v;

    //initVol = v / epsDensity; // Ginger: never change that after birth

    radius = Globals.volume2radius( epsDryMass / epsDensity );

  }

  public double getEpsDryMass() {return epsDryMass;}


  // vfb, trans, 001227: need both to get onto the popup menu....

  public void setDecayR(double d){decayR = d * actualSite.getStepSize();}

  public double getDecayR(){return decayR / actualSite.getStepSize();}


  /** translation: tell some Critter routines to do nothing. In fact,

   *  these are never called. But keep the compiler off our backs... */

  public Move getDesiredMove() {return null;}

  public boolean wantsToReproduce() {return false;}

  public ResourceList getKidShare() {return null;}

  public double getKidDistance() {return 0;}


  // Translation: move and doTheLocomotion translated to BacShove class.


  // Translation:  Diffuse needs to know. Guess this is a reaction id.

  public int getProcess(){return -1;}


  public void die() {
    // almost no biomass left, so get rid of this blob for efficiency reasons
    // no point in converting back using convertBack
    // since death condition is no-biomass-left, book-keeping simpler than Critter....
    remove(); // from Quad tree
    getSpecies().notifyDeath();  // for population
    Globals.log("Death of eps at " + Globals.getCurrentTime());
  }

  public boolean isStarving() {
    return radius <= 0.005;
  }

  // vfb, 991124:  do shoves along the way instead of batch post-process to
  // cut shoving time in half, hopefully
  private void shove() {
    double shoveR;
    double overlapRadius;
    Quad q;
    Quad mobileRoot = Globals.getMobileRoot();
    QuadCircleSpecies neighbor;
    Direction hisDirection;
    shoveR = radius;

    // 1. Shove remainder of quadding tree starting from me. Note that my
    // direction already contains shoves from every neighbor before me in the
    // tree, done by this routine for them.

    shovewin.sphereToBrick(x, y, z, radius);
    q = this.pickWin(shovewin, mobileRoot, false);
    while (q != null) {
      // do one mutual shove
      neighbor = (QuadCircleSpecies) q;
      overlapRadius = shoveR;
      overlapRadius -= Globals.distance( x, y, z,
          neighbor.x, neighbor.y, neighbor.z )
          - neighbor.radius;
      if (overlapRadius > 0) {
        // my side of the shove
        direction.addDirection(x - neighbor.x,
                     y - neighbor.y,
                     z - neighbor.z, overlapRadius);
        // his side of the shove
        hisDirection = ((GetDirectionable)neighbor).getDirection();
        hisDirection.addDirection(neighbor.x - x,
                     neighbor.y - y,
                     neighbor.z - z, overlapRadius);
      }
      // next
      q = q.pickWin(shovewin, mobileRoot, false);
    }

    // 2. check the carrier wall
    if (actualSite.isBiofilm()) {
      // Then get a shove from the concrete wall called carrier as well
      // (none if we don't sit on carrier)
      // carrier is always at the low x side
      overlapRadius = actualSite.getOverlapWithCarrierAt((x - shoveR), y);
      if (overlapRadius > 0) {
        direction.addDirection(1, 0, 0, overlapRadius);
      }
    }

    // 3. apply all the shoves to date (neighbors and carrier)
    double u[] = direction.getWannas();
    x += u[0]; y += u[1]; z += u[2];

    // 4. check boundary conditions
    BacShove.wrapXYZ(this,actualSite);

    // 5. update Mcr
    setMcr();

    // 6. zero out direction--have applied all
    direction.zeroDirection();
  }

  // EPS decay is assumed to be catalyzed by a constant, concn-independent,
  // activity per volume (space, not EPS)
  private void decay() {
    double decayed = decayR * epsDryMass / initVol;
    epsDryMass -= decayed;
    // see LandRoot, in future better not let LandRoot deal with this, not enuf control over which substrate where will get upped
    Globals.getTaxCollector().payTax(x, y, 0, decayed * convertBack);
    // update radius after changing epsDryMass
    radius = Globals.volume2radius(epsDryMass / epsDensity);
  }

  public void step() {
    // Basically, EPS only moves, and that Diffuse's instigation.
    // It's passive stuff. But maybe we want to recycle or age it later.

    long now = Globals.getCurrentTime();
    if ((birthday == now) && (now > 0)) return; // we stepped already

    if (decayR > 0.0) {
      if (isStarving()) { // death if radius very small
        die(); // die does not return?
        return; // Ginger: shouldn't we return here?
      }
      decay();
    }

    // vfb, 991124
    shove();

    // print after update, and only if alive this step
    if ( ( Globals.getCurrentTime() % Globals.getStatCycle() ) == 0 )
      printv();

    getSpecies().notifyBiomass(epsDryMass);
  }

  private static int lastRound = -1;
  private static int lastRun = -1;
  /** printv only does one cell per round */
  public void printv() {
    // most printing is done from Diffuse while the growth step is frozen,
    // here we only print when we try things out
    // some things should be printed for all cells, some not
    // some things should be printed often, some not
    int now = (int) Globals.getCurrentTime();
    int thisRun = Globals.getGeckoDriver().getNumRun();

    // this is printed only for one cell per round
    if ((lastRound != now) || (lastRun != thisRun)) {
      // this is printed at lower frequency ([geckoControl getStatCycle])
      if ( (now % Globals.getStatCycle()) == 0) {

      }

      // this is printed every printstep

      lastRound = now;
      lastRun = thisRun;
    }

    // this is printed for all cells

    // this is printed at lower frequency ([geckoControl getStatCycle])
    if ( (now % Globals.getStatCycle()) == 0) {

    }
  }

  /** Replacement for snapshots, prints all spheres for import into Matlab,
   * print Lattices separately
   */
  private static PrintWriter sphereFile;
  static public void closeSphereFile() {
    if (sphereFile ==  null)
      return;
    sphereFile.flush();
    sphereFile.close();
  }

  public void printSpheresMatlab() {
    int now = (int) Globals.getCurrentTime();
    int thisRun = Globals.getGeckoDriver().getNumRun();

    //String filename = ((BatchUI) Globals.getUI()).getFileStem();
    String filename = new PrintfFormat("eps_%03ld_").sprintf(thisRun);
    filename += new PrintfFormat("%06ld.ml").sprintf(now);

    // do this only once at the beginning of each round
    if ((lastRound != now) || (lastRun != thisRun)) {
      // overwrite--zero out the file
      try {
        sphereFile = new PrintWriter(new FileWriter(filename, false), true);
        sphereFile.close();
      } catch (Exception e) {
        Globals.log("Error : " + e + " Opening output file : " + filename);
      }
      lastRound = now;
      lastRun = thisRun;
    }

    // this time, append instead of overwrite
    try {
      sphereFile = new PrintWriter(new FileWriter(filename, true), true);

      sphereFile.print(new PrintfFormat("%d\t").sprintf(-1)); // species id -1 (process)
      sphereFile.print(new PrintfFormat("%g\t").sprintf(x)); // coords
      sphereFile.print(new PrintfFormat("%g\t").sprintf(y)); // coords
      sphereFile.print(new PrintfFormat("%g\t").sprintf(z)); // coords
      sphereFile.print(new PrintfFormat("%g").sprintf(radius)); // coords
      sphereFile.println();
      sphereFile.close();
    } catch (Exception e) {
      Globals.log("Error : " + e + " Opening output file : " + filename);
    }
  }



  public String toString() {
    return "Eps" + Globals.pointerString(this) +
            "[epsDryMass=" + epsDryMass + "]";
  }



}








