//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator

/** 991209, vfb.
 *  FragStats show species incidence over either sites or an arbitrary
 *  grid laid over the world. If there are N species, the frag plot always
 *  has 2*N + 1 traces: 0 species here, 1 species here, ... n species here,
 *  and speciesA here, speciesB here... Then the Y-axis is a number of
 *  site/grid fragments (say, 8 frags had Hemlock present, 14 frags had only
 *  1 species present, etc.) The X-axis is time. The buckets here are boolean--
 *  each species either is or isn't present at this frag.
 *
 *  This stat makes no sense at all unless you've got a selected group of
 *  species to evaluate. If animal/vegetable/mineral in your scenario, you're
 *  not going to use all species in frag analysis. Hence, can only frag a
 *  composite species, because they can provide the member species list.
 *
 *  The plot is too hard to read with nof lines and species lines, so there
 *  are three modes for fragstats--the nof lines ("frags with 0 species"...),
 *  the species lines, or both.
 */

package gecko.statglue;

import gecko.Globals;
import gecko.*;
import gecko.zoo.*;
import stats.*;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.awt.Color;
import java.util.Vector;

public class FragStatBuckets extends StatSource implements gecko.Steppable {
  public static final int BY_SPECIES = 0;
  public static final int BY_NOF = 1;
  public static final int BY_BOTH = 2;

  private StatBuckets[] stages;
  private int numSpecies;
  private int numFrags;
  // due to initialization order problems, later title changes have no effect....
  private String title = "Fragments With...";
  private Vector fragMcr = null;  // [frag]--these are actually Quads on tree, probably
  private int numGridX = 10;       // unused if frags, else default to 5x5 grid.
  private int numGridY = 10;
  private double xdelta;
  private double ydelta;
  private Species[] species;     // [numspecies]
  private boolean[][] present = null; // opp "absent": [frag][numspecies] 1 or more found
  private Color[] nOfColors = null; // colors for "0 species" etc. traces  : N + 1 OF THEM
  private String compName;
  private int mode = BY_NOF;
  private long lastStep = -10; // not-now = default

  private double[] plottables = null; // derived stats to plot

  public FragStatBuckets(CompositeSpecies speciesList) {
    numSpecies = speciesList.getMembers().size();
    compName = speciesList.getName();
    species = new Species[numSpecies];
    for (int i=0; i<numSpecies; i++) {
      species[i] = (Species)speciesList.getMembers().elementAt(i);
    }
    plottables = new double[(2*numSpecies) + 1];
    nOfColors = new Color[(2*numSpecies) + 1];
    Palette palette = new Palette(numSpecies + 1, Palette.PASTEL);
    for (int i=0; i<numSpecies + 1; i++)
      nOfColors[i] = palette.getColor(i);
    for (int i=0; i<numSpecies; i++) {
      int colorid = species[i].getProgenitor().getColor();
      nOfColors[numSpecies + 1 + i] = QuadDisplay.getColor(colorid);
    }
  }

  /** make sure the derived stuff is in/back in undecided state */
  public void reinit() {
    lastStep = -10; // not-now, ever.
    fragMcr = null;
    numFrags = 0;
    present = null;
  }

  public void setMode(int n) {mode = n % 3;}
  public int getMode() {return mode;}
  public void setNumGridX(int n) {numGridX=n; numFrags=0;}
  public void setNumGridY(int n) {numGridY=n; numFrags=0;}
  /** If using a grid, set up the fragmath. */
  private void setupGrid() {
    title = compName + " grids (" + numGridX + " x " + numGridY + ") With...";
    // and set up the frag grid
    numFrags = numGridX * numGridY;
    fragMcr = null;
    Mcr world = Globals.getWorldMcr();
    xdelta = (world.ux - world.lx) / numGridX;
    ydelta = (world.uy - world.ly) / numGridY;
  }

  public int getNumStats() {return plottables.length;}
  public Color[] getColors() { return nOfColors; }
  public Color getColor(int i) { return nOfColors[i];}
  public String getName(int i) {
    if (i <= numSpecies)
      return i + " species";
    int s = i - (numSpecies + 1);
    return species[s].getName();
  }
  public double getStat(int which) { return plottables[which];}
  public String getTitle() {return title;}


  private int frag(double x, double y) {
    if (fragMcr==null) {
      if (numFrags==0) return 0;
      Mcr world = Globals.getWorldMcr();
      int xid = Math.round( Math.round( Math.floor( (x - world.lx) / xdelta ) ) );
      int yid = Math.round( Math.round( Math.floor( (y - world.ly) / ydelta ) ) );
      if ( (xid<0) || (xid >= numGridX) || (yid<0) || (yid>=numGridY)) {
        return numFrags;
      }
      return xid*numGridY + yid;
    } else {
      for (int i=0; i<numFrags; i++) {
        if ( ((Mcr)fragMcr.elementAt(i)).isInsideOrOn(x, y)) {
          return i;
        }
      }
      return numFrags;
    }
  }

  public void tally(QuadCircleSpecies agent) {
    if (present == null) return; // not counting yet.
    Species aspecies = agent.getSpecies();
    for (int i=0; i<species.length; i++) {
      if (aspecies == species[i]) {
        present[ frag(agent.x, agent.y) ][i] = true;
        return;
      }
    }
  }

  public FragBroker[] getBrokers() {
    int howmany, first;
    if (mode == BY_NOF) {
      howmany = numSpecies + 1;
      first = 0;
    } else if (mode == BY_SPECIES) {
      howmany = numSpecies;
      first = numSpecies+1;
    } else {
      howmany = plottables.length;
      first = 0;
    }
    FragBroker[] brokers = new FragBroker[howmany];
    for (int i=0; i < howmany; i++) {
      brokers[i] = new FragBroker(this, first+i);
    }
    return brokers;
  }

  private void fragHunt() {
    Quad landroot = Globals.getLandRoot();
    Quad aquad = landroot.pickWin(landroot, landroot, false);
    fragMcr = new Vector(10,10);
    while (aquad != null) {
      if (aquad instanceof Siteable) {
        fragMcr.addElement(aquad);
      }
      aquad = aquad.pickWin(landroot, landroot, false);
    }
    /** if have multi-frags, perhaps can dispense with supersite */
    Mcr world = Globals.getWorldMcr();
    if (fragMcr.size() > 1) {
      for (int i=0; i<fragMcr.size(); i++) {
        Mcr frag = (Mcr)fragMcr.elementAt(i);
        if (frag.equals(world)) {
          fragMcr.removeElementAt(i);
          i--;
        }
      }
    }
    fragMcr.trimToSize();
    numFrags = fragMcr.size();
    title = compName + " fragments (" + numFrags + ") With...";
  }

  /** If frags aren't initalized yet, do that first.
   *  Calculate the derived stats.
   */
  public void step() {
    // must only step once per timestep, since reset counters
    if (lastStep == Globals.getCurrentTime())
      return;

    // first make sure we're all initialized, may have to look for frags to use
    // not that present allocates numFrags+1 because some agents are in no frag
    if ( lastStep < 0 ) {
      fragHunt();
      if ((numFrags<2)) {
        fragMcr = null;
        setupGrid();
      }
      present=new boolean[numFrags + 1][numSpecies]; // and all false by language defn.
    }

    // alright, in plottables, 0 through nspecies are the n-of's, rest species
    int nplottable = 2*numSpecies + 1;
    for (int i=0; i<nplottable; i++) {
      plottables[i] = 0;
    }

    // derive the stats for this round into plottable, and reset present flags
    for (int f=0; f<numFrags; f++) {
      int nof = 0;
      for (int s=0; s<numSpecies; s++) {
        if (present[f][s]) {
          nof++;
          plottables[numSpecies + 1 + s]++;
          // having derived, zero out the source
          present[f][s] = false;
        }
      }
      plottables[nof]++;
    }

    lastStep = Globals.getCurrentTime();
  }
}


