//Title:        Gecko Woods redux, model #3
//Version:
//Copyright:    Copyright (c) 2000, All rights reserved
//Author:       Ginger Booth, ginger@gingerbooth.com
//Company:      Yale IBS/CCE
//Description:  "mini-Sortie"
/**
 * 000824, vfb, copied over from Tree to hack into Tree3 model.
 *
 * 000412, vfb, modified height formula to use heightIntercept=1.35 ala
 *              Charlie Canham's latest paper.  Tested and found it made no
 *              difference, so removed it again.
 *
 * 991018, vfb, original Java
 * Base tree of the Tree3s model (miniSortie). Note that the landscape has
 * nothing to do with light competition in Woods. Each step, these agents do
 * their own calculation to find who's shading them. Also note these agents
 * don't do Gecko-style metabolism, having their own rules based on tree
 * field observations.
 *
 * Stats:
 * This species family redefines "Biomass" as "Basal Area", including renaming
 * the Globals.XXX_PLOT_NAMEs. If adding other species families, this could become
 * a problem. Note that biomassTransfers and biomassTheft are nulls--this species
 * family only does observeBiomass and observeBiomassGrowth.
 */

package gecko.woods;
import gecko.*;
import gecko.zoo.*;
import gecko.statglue.*;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.awt.*;
import java.util.Random;

public class Tree3 extends QuadCircleSpecies implements StageObservable, Biomassable, ReapableTree {
  protected Tree3Params params; // cut out of this object to save memory

  // The working settings for the species include x,y,z,radius,color in
  // QuadCircleSpecies. We're storing height in z. radius is my visible CROWN radius.
  protected double stemDBH;           // my trunk diameter, in cm--relates to recruitRadius
  protected double saveDBH;           // last value
  protected double crownBottom;       // z of crown bottom (z is crown top)
  protected double nutrientFactor;    // weighted average ba/unit area of sites on top of
  // these are our competitive losses for this year over footprint
  // originally implemented as total, top, bottom, side, but decided to remove 091100
  protected double competeLoss;       // total % lost to competition over footprint

  // don't allocate these over and over....
  protected static McBrick crossBrick = new McBrick();
  protected static McBrick crownBrick = new McBrick();
  protected static McrCircle stem = new McrCircle();

  // we are going to want to keep track of which direction to grow in.
  // not an ability supplied by QuadCircleSpecies.
  protected Direction cropWannago;  // scratchpad for where we're headed
  protected Direction nutrientWannago;  // scratchpad for where we're headed
  protected RunningAverage growthMemory; // delta-dbh memory
  protected boolean dieASAP = false; // more direct way of flagging....

  protected final static double sqrt1Over3 = Math.sqrt(1.0/3.0);

  static {
    // since "Biomass" is unclear, rename Global plots
    Globals.BIO_PLOT_NAME        = "Basal Area";
    Globals.BIO_GROWTH_PLOT_NAME = "Growth vs. Basal Area";
    GrassResources resources = (GrassResources) Globals.getResources();
    resources.setNumResources(3);
    resources.setBioScale(0,1.0);  resources.setResourceName(0,"null");
    resources.setBioScale(1,1.0);  resources.setResourceName(1,"tree basal area");
    resources.setBioScale(2,1.0);  resources.setResourceName(2,"site quality");
  }

  public boolean stageObservableCheck() {return true;}

  /** Shall I die? Internally driven choice. "Die" is an order.
   *  Growth-dependent mortality and per-species random death. For the
   *  nonce, we die Now, we vanish Now. Later, want to make the hulks
   *  stick around a while?
   */
  public boolean shallIDie() {
    // species-specific random death? apply before full-sun escape clause
    // use Woods global to turn on/off so don't need to screw with all the
    // Species definitions to turn off.
    if (dieASAP)
      return true;

    if ( Woods.isSpeciesRandomDeath() ) {
      if ( Globals.getRandomSource().nextDouble() < params.mortRandom ) {
        return true;
      }
    }

    // 000912, vfb: use min radial increment on dbh instead as culling factor
    if ( growthMemory.getAverage() < 2.0 * params.minRadialIncrement) {
      return true;
    }

    // if pushed offsite, have no nutrients left, could have height 0....
    if ( z == 0)
      return true;

    return false;
  }

  /** Logically, this belongs in the Reapers. However, since more than one
   *  Reaper needs it, I'm putting it here, so the Tree3 knows how to take
   *  damage, instead of the Reaper knowing the mechanics of inflicting it.
   *
   *  The crown is reduced in volume by the damageRate. Written for hurricanes.
   *  It's the radius that gets reduced, not the crown depth.
   */
  public void crownDamage(double damageRate) {
    double crownH = z - crownBottom;
    double vol = params.coneVolume(radius, crownH);
    vol *= 1.0 - damageRate;
    radius = params.coneRadius(vol, crownH);
  }
  /** Logically, this belongs in the Reapers. However, since more than one
   *  Reaper needs it, I'm putting it here, so the Tree3 knows how to take
   *  damage, instead of the Reaper knowing the mechanics of inflicting it.
   *
   *  The crown is reduced in volume by the damageRate, but only from the
   *  bottom, up to height.
   */
  public void crownDamageBottom(double damageRate, double height) {
    if (height < crownBottom) return;
    double depthAffected = height - crownBottom;
    // to take damageRate percent off volume, without changing radius,
    // you simply raise the bottom.
    double newDepth = depthAffected * (1.0-damageRate);
    crownBottom = height - newDepth;
  }

  /** ReaperOld wants to know, but we have no basis in params to say, so
   *  have an arbitrary answer in this model. Was using variants here
   *  to examine effect on ba-growth of culling old instead of young.
   */
  public boolean isFullGrown() {
    return z > (0.9 * params.maxHeight);
  }

  /** An order from without--"You Die." aot "I die", at my choice. */
  public void die() {
    remove();  // from QuadTree / world
    Species s = getSpecies();
    s.notifyDeath();  // for population
    s.notifyDeath(this); // for stage
    s.notifyBioTrans(BiomassTransfers.LOSS_STARVE, getBiomass()); // for biomass
  }


  /** use a radius equivalent to coneVolume for cylinder */

  public double getCompetitiveRadius() {

    return radius * sqrt1Over3;

  }

  /** use a radius equivalent to coneVolume for cylinder */

  static public double getCompetitiveRadius(double radius) {

    return radius * sqrt1Over3;

  }


 /** Bask = my year in the sun. Kill seedlings that get too close.
   * Crown competition with neighbors.
   */
  public void bask() {
    // constructs we'll use for picking and competing.
    crownBrick.circleToSquare(x, y, getCompetitiveRadius());
    crownBrick.lz = z - crownBottom;
    crownBrick.uz = z;
    stem.circleToSquare(x, y, Woods.getThinningFactor() * stemDBH / 200);

    // check for limiting resource and set nutrientFactor
    findNutrientFactor();

    Quad plantRoot = Globals.getPlantRoot();
    // only consider plants after me--plants before me have already competed
    // with me.
    Quad somequad = this.pickWin(crownBrick, plantRoot, false);
    Tree3 competitor;
    double crownVol = crownBrick.getVolume();
    while (somequad != null) {
      if ((somequad instanceof Tree3) && (somequad != this)) {
        boolean skipCrown = false;

        // first issue--crown aversion
        competitor = (Tree3) somequad;
        crossBrick.circleToSquare(competitor.x, competitor.y, competitor.getCompetitiveRadius());
        crossBrick.lz = competitor.crownBottom;
        crossBrick.uz = competitor.z;
        double competitorVol = crossBrick.getVolume();
        if ( crownBrick.intersectMcb(crossBrick, crossBrick) != Mcr.MCR_DISJOINT ) {

           // second issue--trunk aversion
          if ((competitor.getAge() < competitor.params.recruitAge)) {
            if (getAge() > params.recruitAge + 4) {
              if (stem.isInside(competitor.x, competitor.y)) {
                competitor.die();
                skipCrown = true;
              }
            }
          }

          if (!skipCrown) {
            // the volume of the intersection, footprint area, reference column volume
            double crossVol = crossBrick.getVolume();
            double crossArea = crossBrick.getArea();
            double contestTop = z > competitor.z ? z : competitor.z;
            double contestBottom = crownBottom < competitor.crownBottom ? crownBottom : competitor.crownBottom;
            double areaVol = crossArea * (contestTop - contestBottom);

            // maximum percent loss is crossVol/areaVol, because our losses are in footprint
            double maxLoss = (crossVol / areaVol ) * crossArea;

            // now divide up that loss based on TOTAL volume * light coeff--compet. push
/*
            double myWeight= crownVol * (1.0 - params.lightCoeff);
            double hisWeight = competitorVol * (1.0 - competitor.params.lightCoeff);
*/

            double myWeight;
            double hisWeight;
            if (z < competitor.z) {
              myWeight = competitor.z;
              hisWeight = z * (1.0 - params.lightCoeff);
            } else {
              myWeight = competitor.z * (1.0 - competitor.params.lightCoeff);
              hisWeight = z;
            }

            // note += accumulating through every competition!
            double myLoss = maxLoss * (myWeight / (myWeight + hisWeight));
            competeLoss += myLoss;
            double hisLoss = maxLoss * (hisWeight/(myWeight + hisWeight));
            competitor.competeLoss += hisLoss;
/*
double myLoss;
double hisLoss;
myLoss = 0.5 * (1.0 - competitor.params.lightCoeff) * maxLoss;
hisLoss = 0.5 * (1.0 - params.lightCoeff) * maxLoss;
competeLoss += myLoss;
competitor.competeLoss += hisLoss;
*/

            cropWannago.addDirection(x-competitor.x, y-competitor.y, myLoss);
            competitor.cropWannago.addDirection(competitor.x-x, competitor.y-y, hisLoss);
          }
        }
      }
      somequad = somequad.pickWin(crownBrick, plantRoot, false);
    }
  }

  /** NOT to be confused with findNutrientFactor--this just for comms with params. */
  public double getNutrientFactor() {return nutrientFactor;}

  /** another subroutine of bask. Gets area covered by site, basically.
   */
  protected void findNutrientFactor() {
    // this lightlevel is per unit area of equiv area circle. Use same to
    // pick on the landscape.
    double nutrientArea = 0; // none until find some....

    Quad landRoot = Globals.getLandRoot();
    Quad somequad = landRoot.pickWin(crownBrick, landRoot, false);
    LimitResourceQuad limiter;

    // if agent is at the edge of one site, and another site doesn't compensate,
    // then need to increase competeLoss and cropWannago. But can't tell whether
    // to tally into those until we've considered all sites.
    nutrientWannago.zeroDirection();
    double loss = 0;
    double edgeLosses = 0;
    while (somequad != null) {
      if (somequad instanceof LimitResourceQuad) {
        limiter = (LimitResourceQuad) somequad;
        if (crownBrick.intersectMcr(limiter, crossBrick) != Mcr.MCR_DISJOINT) {
          nutrientArea += limiter.getQuality(x, y) * crossBrick.getArea();
          if (crossBrick.getArea() < crownBrick.getArea()) {
            loss = limiter.getQuality(x, y) * (crownBrick.getArea() - crossBrick.getArea());
            edgeLosses += loss;
            nutrientWannago.addDirection(crossBrick.getCenterX() - x,
              crossBrick.getCenterY() - y, loss);
          }
        }
      }
      somequad = somequad.pickWin(crownBrick, landRoot, false);
    }

    nutrientFactor = nutrientArea/crownBrick.getArea();
    // the following fix is occassioned by crowns being capable of getting
    // sufficiently tiny (Ex10-16 or so) that multiplying them goes NaN.
    // this prevents NaN from propagating, and increases the chance that the
    // un-Tree3 dies very very soon.
    if (Double.isNaN(nutrientFactor))
      nutrientFactor = 0;

    // let's dig down into the wanna's on direction, because we want only a
    // significant vector to matter.
    double signif = .05; // less than 5%, consider roundoff.
    loss = nutrientWannago.getMagnitude2d();
    if (loss > signif * edgeLosses) {
      competeLoss += loss;
      cropWannago.addDirection(nutrientWannago.wannax, nutrientWannago.wannay, loss);
    }
  }

  // these weird utilities to clarify grow
  private double oneNewDBH(double oldVol, double oldDBH, double budget, double z) {
    double mRBH = Math.sqrt( (oldVol + budget) / (Math.PI * z) );
    double dbh = mRBH*200; // return cm, diam.
if (Double.isNaN(dbh))
System.err.println("oneNewDBH went NaN, oldvol=" + oldVol + ", oldDBH=" + oldDBH
+ ", budget=" + budget + ", z=" + z);
    return dbh;
  }
  private double oneNewZ(double beginZ, double oldDBH, double deltaDBH) {
    double dbhZ = params.zFormula(oldDBH);
    double newDbhZ = params.zFormula(oldDBH + deltaDBH);
    // z is affected by more than the dbh-based z formula
    double answer = beginZ + (newDbhZ - dbhZ);
if (Double.isNaN(answer))
System.err.println("oneNewZ went NaN");

    return answer;
  }

  /** Stem growth is no longer related to Sortie. Grow is responsible for
   *  moving and growing the stem and crown. And resetting whatever is derived
   *  from that.
   */
  public void grow() {
    // the bask competition set competeLoss scratchpad, which
    // must be zeroed by the end of grow.
// System.err.println("time " + Globals.getCurrentTime() + this);

    // my change in crownRadius determines how far I move
    double saveCrownRadius = radius;
    saveDBH = stemDBH;
    double saveCrownDepth = z - crownBottom;
    double saveCompetitiveRadius = getCompetitiveRadius();
    double crownDepth = saveCrownDepth; // crownBottom is the kept state

    // let's shrink, then grow

    // step 1: crop -current- crown before growing it
    // this gets ugly in that we have competitive losses in -area- but will
    // apply them to -volume-. Ugh.
    double carea = Math.PI * saveCompetitiveRadius * saveCompetitiveRadius;
    if (competeLoss > carea) {
      competeLoss = carea;
      dieASAP = true;
      return;
    }
    double crownVolWas = params.coneVolume(saveCrownRadius, saveCrownDepth);
    double crownVolNew = crownVolWas * (1.0 - ( competeLoss / carea ));

    // now the tricky bit--side losses are applied to cone base term,
    // then the height changed derived from there.
    double newRBasis = ( (carea - competeLoss) / carea ) * (Math.PI / 3.0) * ( radius * radius );
    radius = Math.sqrt( 3.0 * newRBasis / Math.PI );

    if (z <= 0) {
      dieASAP = true;
      return;
    }

    // step 2: find budget of new biomass to partition into crown
    // and stem (crown + stem != full budget, that includes roots,
    // and the crown budget is m^3 of a different density)
    double stemRM = stemDBH/200;
    double oldMidZ = z - crownDepth/2.0;
    double saveTrunkVol = Math.PI * stemRM * stemRM * oldMidZ;

    // fullBudget is in m^3 of this-density
    double effArea    = carea - competeLoss;
    if (effArea < 0){
      effArea = 0;
      competeLoss = carea;
      growthMemory.addMemory(0.0);
      return;
    }
    double fullDepthCrown = z * params.crownDepthToHeight;
    double percentDepth = saveCrownDepth / fullDepthCrown;
    if (percentDepth > 0) {
      effArea *= percentDepth;
    }

    double fullBudget = params.getGrowthBudget(this, effArea);
    // stemBudget is in m^3 of this-density.
    double stemBudget = params.getStemPart(this, fullBudget);
// System.err.println("stemBudget=" + stemBudget + ", trunkVol=" + saveTrunkVol + ",oldMidZ=" + oldMidZ);

    // step 3: decide how to sheathe the stem with new biomass.
    // I don't know how to solve the delta-r/delta-z problem directly,
    // so do it piecemeal. Also, it is the mid-crown height we go by
    // for stem volume, not the height, so....
    double newZ = oldMidZ;
    double newDBH = stemDBH;
    // n rounds, getting successively closer to a good answer
    int maxRounds = 20;
    double lastCalcDBH = saveDBH;
    double lastCalcZ = z;
    double diffDBH, diffZ; // percentDiff
    double maxDiff = .0001; // 4 digits accuracy
    convergence:
    for (int i=0; i<maxRounds; i++) {
      newDBH = oneNewDBH(saveTrunkVol, saveDBH, stemBudget, newZ);
if (Double.isNaN(newDBH)) {
System.err.println("stem went NaN, exit");
dieASAP = true;
return;
}
      newZ   = oneNewZ(oldMidZ, saveDBH, (newDBH - saveDBH));
      diffDBH = Math.abs(newDBH - lastCalcDBH) / lastCalcDBH;
      diffZ = Math.abs(newZ - lastCalcZ) / lastCalcZ;
      if ( (diffDBH < maxDiff) && (diffZ < maxDiff)) {
        break convergence;
      }
      lastCalcDBH = newDBH;
      lastCalcZ = newZ;
// System.err.println("round" + i + ", newDBH=" + newDBH + ",newZ=" + newZ);
    }

    // convert back to height--were working with mid-crown z
    double deltaZ = newZ - oldMidZ;
    z += deltaZ;
    crownDepth += (deltaZ) * params.crownDepthToHeight;
    crownBottom = z - crownDepth;
    growthMemory.addMemory(newDBH - saveDBH);
    stemDBH = newDBH;

    // step 4: use linear growth on crown
    radius += params.crownToStem * (stemDBH - saveDBH);
if (Double.isNaN(stemDBH)) {
System.err.println("stem went NaN, exit");
dieASAP = true;
return;
}
if (Double.isNaN(radius)) {
System.err.println("radius went NaN");
dieASAP = true;
return;
}

    // step 4: pay for leaf turnover on cropped crown (proportioned to fine
    // root turnover...) and add leafBudget to get new radius. CrownDepth
    // has changed as much as it's going to by now.
    /*
    double leafVolumeWas = params.coneVolume(radius, saveCrownDepth);
    double leafLoss = leafVolumeWas * params.leafTurnover;
    double leafVolumeNew = leafVolumeWas - leafLoss + leafBudget;
    double newRadius = params.coneRadius(leafVolumeNew, crownDepth);
    radius = newRadius;
/*
System.err.println("leaf vol was=" + leafVolumeWas + ", new=" + leafVolumeNew
+ ", loss=" + leafLoss + ", budget=" + leafBudget + ", radius=" + radius
+ ", from=" + saveCrownRadius);
*/

    // step 5: move by the amount the competitive radius changed
//    double dR = Math.abs(getCompetitiveRadius() - saveCompetitiveRadius);
    double dR = Math.abs(radius - saveCrownRadius);
    double[] unitv = cropWannago.getDirection();
    x += unitv[0] * dR;
    y += unitv[1] * dR;
    cropWannago.zeroDirection();

    setMcr();

    // clear for next year
    competeLoss = 0;
  }

  /** Reproduce doesn't implement Sortie's model, though it's sort of following
   *  the spirit of it.
   */
  public void reproduce() {
    Random rands = Globals.getRandomSource();
    double distance;

    // 0. How many? Use whole part of answer as a given, fractional part as probability.

    // fecundity proportional to linear dimension cubed? with bigFertility
    // defined as fertility at maxHeight?
//    double dnumKids = params.bigFertility * Math.pow((z/params.maxHeight), 2.5);
//    double dnumKids = params.bigFertility * Math.pow((z/params.maxHeight), 3);
double dnumKids = params.bigFertility * Math.pow(stemDBH/75, 2.5);

    // Sortie paper divides by 100 cm, comparing tree to an absolute.  They also
    // made it proportional to square instead of cube....
    // double dnumKids = params.bigFertility * stemDiameter * stemDiameter / 10000;

    int numKids = Math.round(Math.round(Math.floor(dnumKids)));
    dnumKids -= numKids;
    if (rands.nextDouble() < dnumKids)
      numKids++;
    for (int i=0; i<numKids; i++) {

      // 1. for each, how far? Since we're not using quadrats like Sortie, any
      // discrete (a la Poisson) distribution is meaningless--we want a continuous distance.
      // tests show the mean of this thing is slightly less than the mean desired.
      // like on 10,000 trials, mean 10 desired -> mean 9.97 actual. Close enuf.

      distance = (params.recruitMeanD * Math.exp(rands.nextDouble()) ) / (Math.E - 1.0);

      // 2. OK, make a kid.
      Speciesable lbaby = makeKid();
      Tree3 baby = (Tree3) lbaby;
      double direx = 2.0 * Math.PI * rands.nextDouble();
      baby.x = x + ( distance * Math.cos(direx) );
      baby.y = y + ( distance * Math.sin(direx) );

      // do this before wrap-killing or population gets out of whack
      getSpecies().notifyBirth();

      Mcr limits = Globals.getWorldMcr();
      double sizex = limits.ux - limits.lx;
      double sizey = limits.uy - limits.ly;

      // 3. possibly stillborn, or wrapped into world
      if (baby.x > limits.ux) {
        if (Woods.isWrapX()) {
          baby.x = ((baby.x - limits.lx) % sizex) + limits.lx;
        } else {
          baby.die();
          return; // don't put baby on tree
        }
      } else if (baby.x < limits.lx) {
        if (Woods.isWrapX()) {
          baby.x = ((baby.x - limits.lx) % sizex) + sizex;
          baby.x = (baby.x % sizex) + limits.lx;
        } else {
          baby.die();
          return; // don't put baby on tree
        }
      }
      if (baby.y > limits.uy) {
        if (Woods.isWrapY()) {
          baby.y = ((baby.y - limits.ly) % sizey) + limits.ly;
        } else {
          baby.die();
          return; // don't put baby on tree
        }
      } else if (baby.y < limits.ly) {
        if (Woods.isWrapY()) {
          baby.y = ((baby.y - limits.ly) % sizey) + sizey;
          baby.y = (baby.y % sizey) + limits.ly;
        } else {
          baby.die();
          return; // don't put baby on tree
        }
      }

      // 4. Although logically, baby would die now if fell too close to
      // someone else's trunk, we leave it to the elder to kill the baby.
      // hang on quad tree
      baby.setMcr();
// System.err.println("new baby: " + baby);
      Globals.getPlantRoot().addNewbie(baby);
    }
  }

  public StageObserver getStageObserver() {
    if (getSpecies().getHoldData() == null)
      makeHeightObserver();
    return (StageObserver) getSpecies().getHoldData();
  }

  /** The activities of a Tree3 per step. One step = one year. */
  public void step() {
    // added to tree -this- step--skip out
    if ( (Globals.getCurrentTime() > 0) && (getAge() == 0))
      return;
      
    getSpecies().notifyAlive(this); // for stages

    // if < recruitAge, this tree is still growing to recruitRadius,
    // and does nothing active in the meantime. It is alive....
    if ( (Globals.getCurrentTime() >= params.recruitAge) && (getAge() < params.recruitAge)) {
      // nah, isn't included in basal area, either, since not right yet.
      // getSpecies().notifyBiomass(getBiomass());
      return;
    }

    bask();
    grow();
    if (shallIDie()) {
      die();
      return;
    }
    reproduce();
    getSpecies().notifyBiomass(getBiomass());
  }

  /** Tree3 uses basal area as biomass. */
  public double getBiomass() {
    double r = stemDBH / 2.0;
    double cm2 = Math.PI * r * r;
    return cm2 / 10000.0; // return m^2 instead of cm^2
  }

  /** Try W = c d^2.5 for biomass ala Landsberg 000711. */
  /*
  public double getBiomass() {
    double cw = 5700; // units are kg?... don't cancel..."reasonable value"
    double mass = cw * Math.pow(stemDiameter, 2.5);
    return mass;
  }
  */
  /** Is R sigmoidal? Only really useful with single tree...*/
  /*
  public double getBiomass() {
    return stemDiameter/2.0;
  }
  */

  public double getCrownBottom() {return crownBottom;}


  /** Make a gizmo for interpreting tree sizes for plots. */

  public void makeHeightObserver() {
    HeightStageObserver hso = new HeightStageObserver();
    hso.setNumStages(3);
    hso.setStageMaxHeight(params.maxHeight);
    // heightInitSlope is in m/cm
    double minHeight = params.zFormula(params.recruitDBH);
    hso.setStageMinHeight(minHeight);
    getSpecies().setHoldData(hso);
  }

  /** Paint myself into the world so I can show shadow (height, sortof) */
  public void paint(QuadDisplay qd) {
    final int minCircle = 5; // if circles < this, draw a rect
    final int minTrunk = 10; // if circle < this, skip trunk
    Graphics graphics = qd.getOffscreen();
    double squareR = McrCircle.SqrtPiOver2 * radius;
    int pc = qd.w2p(2 * squareR); // if circle < minCircle pixels, speed it up.
    synchronized (graphics) {
      if (pc < minCircle) {
        // don't waste time treating it as a circle--also draws smaller
        qd.drawRect(qd.getColor(color), x - squareR, y - squareR, x + squareR, y + squareR, true);
        return;
      }
      if (pc < minTrunk) {
        // skip the trunk nonsense
        qd.drawCircle(color, x, y, radius, false);
        return;
      }
      qd.drawCircle(color, x, y, radius, false);
      double stemR = 0.5*stemDBH/100.0;
      pc = qd.w2p(2 * stemR * McrCircle.SqrtPiOver2); // if trunk < 4 pixels, speed it up.
      if (pc < minCircle) {
        stemR *= McrCircle.SqrtPiOver2;
        qd.drawRect(qd.getColor(color), x - stemR, y - stemR, x + stemR, y + stemR, true);
      } else
        qd.drawCircle(color, x, y, stemR, true); // trunk
      if (Woods.isShowHeights()) {
        String label = "" + TruncVal.truncVal(z,3);
        double shortR = 0.7*radius;
        qd.drawString(qd.getColor(color), label, x - shortR, y + 2*stemR, x+shortR, y+radius);
      }
    }
  }

  // =======================================================================
  // routines below this line have to do with OOP practice, mechanics of
  // getting born into the world, etc. Little logic regarding species behavior.
  // =======================================================================

  public Tree3() {
    this(new Tree3Params());
  }

  public Tree3(Tree3Params tp) {
    super();
    setParams(tp);
    biomassId = 1; // report me as basal area for biomass, resource id 1
    setType(Quad.QUAD_DRAWMYSELF); // tells quad display not to draw me.
    cropWannago = new Direction();
    nutrientWannago = new Direction();
    growthMemory = new RunningAverage(params.yearsForMRI);
  }

  protected void setParams(Tree3Params tp) {params = tp;}
  protected Tree3Params getParams() {return params;}

  public String toString() {
    final int sigdigits = 5;
    return getSpecies().getName() + Globals.pointerString(this) + "[" +
      "height(m)=" + TruncVal.truncVal(z, sigdigits) +
      ",stemD(cm)=" + TruncVal.truncVal(stemDBH, sigdigits) +
      ",crownR(m)=" + TruncVal.truncVal(radius, sigdigits) +
      ",crownDepth(m)=" + TruncVal.truncVal(z-crownBottom, sigdigits) +
      ",x=" + TruncVal.truncVal(x, sigdigits) +
      ",y=" + TruncVal.truncVal(y, sigdigits) +
      ",birthday=" + birthday
    ;
//      ",super=" + super.toString();
  }

  public void setFrom(Tree3 other) {
    super.setFrom(other);
    params = other.params; // redundant, but better safe than sorry....
    // defining settings handled by Tree3Params params
    // here we need only worry about working settings and initialization, as needed.
    // 000627: I'm not sure why this is here instead of in makeKid, but while I'm
    // do a huge mechanical blitz, hardly the time to introduce side effects....
    nutrientFactor = 1.0;
    // we're telling species to hold a StageObserver for us--
    // if none, make one now, since we've clearly got the info
    // by now to make one.
    if (getSpecies().getHoldData() == null) {
      makeHeightObserver();
    }
  }

  public Object clone() {
    return makeKid();
  }

  public Speciesable makeKid() {
    // make sure the derived parameters are set up
    params.setDerived();
    Tree3 kid = new Tree3(params);
    kid.setFrom(this);
    kid.stemDBH = params.recruitDBH; // their term for initRadius
    // must initialize radius & crownbottom before sync, because sync is incremental
    kid.radius = params.initCrownRadius;
    kid.z = params.initHeight;
    // chicken and egg problem--syncWithStem sets z first time, so must correct
    // its conclusions re the crownBottom.
    kid.crownBottom = kid.z * (1 - kid.params.crownDepthToHeight);
    kid.setBirthday(Globals.getCurrentTime());
    getSpecies().notifyBirthToMom(this);
    return (Speciesable) kid;
  }

  // order matters! see setMaxCrownRadius, etc. below.
  // left off a number of properties we don't generally vary
  protected static String[] props = { "maxHeight", "initHeight", "height50",
    "crownDepthToHeight", "lightCoeff",
    "recruitMeanD", "bigFertility",
    "crownToStem", "initCrownRadius", "minRadialIncrement",
    "woodDensity", "stemPartBegin", "stemPartEnd", "mortRandom" };
  public String[] getPropertyNames() {return props;}
  public double getMaxHeight(){return params.maxHeight;}
  public void setMaxHeight(double v) {params.maxHeight = v;}
  public double getInitHeight(){return params.initHeight;}
  public void setInitHeight(double v) {params.initHeight = v;}
  public double getHeight50(){return params.height50;}
  public void setHeight50(double v) {params.height50 = v;}

  // not on ui at present--rarely change
public int getYearsForMRI(){return params.yearsForMRI;}
public void setYearsForMRI(int v) {params.yearsForMRI = v;}


  public double getInitCrownRadius(){return params.initCrownRadius;}
  public void setInitCrownRadius(double v) {params.initCrownRadius = v;}
  public double getWoodDensity(){return params.woodDensity;}
  public void setWoodDensity(double v) {params.woodDensity = v;}
  public double getStemPartBegin(){return params.stemPartBegin;}
  public void setStemPartBegin(double v) {params.stemPartBegin = v;}
  public double getStemPartEnd(){return params.stemPartEnd;}
  public void setStemPartEnd(double v) {params.stemPartEnd = v;}
  public void setCrownToStem(double v) {params.crownToStem = v;}
  public double getCrownToStem(){return params.crownToStem;}

  public double getCrownDepthToHeight(){return params.crownDepthToHeight;}
  public void setCrownDepthToHeight(double v) {params.crownDepthToHeight = v;}
  public double getMinRadialIncrement(){return params.minRadialIncrement;}
  public void setMinRadialIncrement(double v) {params.minRadialIncrement = v;}

  public double getLightCoeff(){return params.lightCoeff;}
  public void setLightCoeff(double v) {params.lightCoeff = v;}
  public double getMortRandom(){return params.mortRandom;}
  public void setMortRandom(double v) {params.mortRandom = v;}
  public double getRecruitMeanD(){return params.recruitMeanD;}
  public void setRecruitMeanD(double v) {params.recruitMeanD = v;}
  public double getBigFertility(){return params.bigFertility;}
  public void setBigFertility(double v) {params.bigFertility = v;}
  public double getRecruitDBH(){return params.recruitDBH;}
  public void setRecruitDBH(double v) {params.recruitDBH = v;}
  public int getRecruitAge(){return params.recruitAge;}
  public void setRecruitAge(int v) {params.recruitAge = v;}

}

