//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 000128, vfb, original
 * I need to rework the logic of ScaredyGrazer and OferStateGrazer so I can
 * read it better. They've become way too complex. So this is a helper class
 * that contains the code for scoping out the situation: current best plant,
 * refuge, threat, etc.
 */


package gecko.zoo;
import gecko.*;
import java.util.Random;
import courseware.jfcdep.TruncVal;

public class ScaredySituation {
  private ScaredyGrazer viewpoint;
  // these variables are accessible at the package level because this is
  // really substructure of ScaredyGrazer--let's not make it fat and slow
  // with access routines....
  Critter threat;
  // note that these distances are EDGE TO EDGE, and can often negative, meaning
  // that viewpoint's edge is INSIDE the other
  double threatD; // distance edge to edge, viewpoint to threat
  Critter yumPlant;  // best to-eat right now
  double yumD; // distance edge to edge, viewpoint to yumplant
  Critter farYumPlant; // best to-eat within smell radius
  double farYumD; // distance edge to edge, viewpoint to faryumplant
  Critter safePlant;   // always found by smell radius--closest
  double safeD; // distance edge to edge, viewpoint to safeplant
  boolean amHidden; // probably by safePlant if so, but possibly more than one
  int threatCount; // set to overall number of spiders I see (threat=closest)
  double normalBiteR;
  double refugeBiteR;


  public ScaredySituation() { }

  /** After this routine is called, all the variables of this class are
   *  correctly set for this viewer, now.
   *  Not much point looking for spiders if I'm fearless, so that yields
   *  "I see no threats".
   */
  public void lookAround(ScaredyGrazer viewer, boolean fearless) {
    viewpoint = viewer;
    threat = null;
    yumPlant = farYumPlant = safePlant = null;
    threatD = yumD = farYumD = safeD = 0.0;
    // eggs are oblivious
    if (viewpoint.mood == ScaredyGrazer.SCAREDY_EGG)
      return;

    amHidden = viewpoint.amHidden();
    threatCount = 0;

    // look at spiders first, because if I'm stupid, I'll let the
    // threat situation change my mind about yumPlant.... Old version
    // ScaredyGrazer did choose refuge to eat when preferred was available.
    if (!fearless)
      lookForSpiders();
    lookAtPlants();
  }

  /** Debug **/
  public void lookReport() {
    final StringBuffer buffy = new StringBuffer(255);
    buffy.append("t" + Globals.getCurrentTime() +
      ", goppper(" + TruncVal.truncVal(viewpoint.x) +
      "," + TruncVal.truncVal(viewpoint.y) +
      ")");
    if (yumPlant == null)
      buffy.append(", no yumplant");
    else
      buffy.append(", yum type " + yumPlant.getSpecies().getName() +
        ", dist " + TruncVal.truncVal(yumD));
    if (safePlant == null)
      buffy.append(", no safeplant");
    else
      buffy.append(", safeD=" + TruncVal.truncVal(safeD));
    if (threat == null)
      buffy.append(", no threat");
    else
      buffy.append(", threatD=" + TruncVal.truncVal(threatD) +
        " at (" +  TruncVal.truncVal(threat.x) +
      "," + TruncVal.truncVal(threat.y) + ")");
    System.err.println(buffy);
  }

  private void lookAtPlants() {
    Quad pq;
    // the any's don't meet my criteria, so only get put in the class vars
    // if I can't find anything better for that class of plant
    Critter plant, anyYum, anyFarYum;
    anyYum = anyFarYum = null;

    // OK, old ScaredyGrazer looked for safePlants (big enough to hide) and
    // prey plants (only big enough to eat) separately. So need to find refugeLunch
    // by different criteria than safePlant.
    Critter refugeLunch = null;

    Random rand = Globals.getRandomSource(); // for multiple satisfactories
    Quad plantRoot = Globals.getPlantRoot();

    normalBiteR = viewpoint.calcBite(viewpoint.getGrowRate());
    refugeBiteR = viewpoint.calcBite(viewpoint.getRefugeGrowRate());

    double farR = viewpoint.getRadius() * viewpoint.getRefugeRadius();
    Species refuge = viewpoint.getRefuge();

    boolean isRefuge;
    boolean isBigEnough;
    boolean isClose;
    boolean canSmell;
    double thisD;
    safeD = Double.POSITIVE_INFINITY;

    // OK, look at all plants within farR, set variables as we go
    pq = plantRoot.pickRadius(farR, viewpoint.x, viewpoint.y, plantRoot);
    if (pq != null) do {
      plant = (Critter) pq;
      if (viewpoint.attacks(plant)) {

        // let's collect booleans, then apply, otherwise the if/else
        // nesting gets awfully hard to follow
        isRefuge = plant.getSpecies() == refuge;
        if (isRefuge)
          isBigEnough = plant.getRadius() > viewpoint.radius;
        else
          isBigEnough = plant.getRadius() > normalBiteR;
        thisD = Globals.distance(plant.x, plant.y, viewpoint.x, viewpoint.y)
                 - plant.radius - viewpoint.radius;
        isClose = thisD < 0;
        canSmell = thisD + viewpoint.radius < viewpoint.radius * viewpoint.getSmellRadius();

        // alright, so let's classify this plant as one or more of our targets
        if (isRefuge) {
          // three cases:
          // 1. no refuge found yet, any is better than none.
          // 2. big enough refuge already, only closer & bigenuf is better
          // 3. too-small refuge found at dist, any bigger is better

          if ( (safePlant == null) ||
               ( (safePlant.radius > viewpoint.radius) && (isBigEnough) && (thisD < safeD) ) ||
               ( (safePlant.radius <= viewpoint.radius) && (plant.radius > safePlant.radius) )
             )
          {
            safePlant = plant;
            safeD = thisD;
          }

          // separate question--is this a refuge lunch contender?
          isBigEnough = plant.radius > refugeBiteR;
          boolean wasBigEnough = refugeLunch != null ? refugeLunch.radius > refugeBiteR : false;
          if (isClose) {
            if (refugeLunch == null)
              refugeLunch = plant;
            else if ( ( (wasBigEnough) && (isBigEnough) && (rand.nextDouble() > 0.5) ) ||
                      ( !(wasBigEnough) && (plant.radius > refugeLunch.radius) ) )
              refugeLunch = plant;
          }

        } else { // not refuge
          // yumPlant contender--want close, want far, want any, so more cases
          if (isClose) {
            if (isBigEnough) {
              if ( (yumPlant==null) || ( (yumPlant != null) && (rand.nextDouble() > 0.5) ) )
                yumPlant = plant;
            } else { // not bigEnough
              if ( (anyYum==null) || ( (anyYum != null) && (plant.radius > anyYum.radius) ) ) {
                anyYum = plant;
              }
            }
          } else { // not close
            if (canSmell) {
              if (isBigEnough) {
                if ( (farYumPlant==null) ||
                     ( (farYumPlant != null) && (thisD < farYumD) )
                   )
                {
                  farYumPlant = plant;
                  farYumD = thisD;
                }
              } else { // not bigEnough
                if (  (anyFarYum == null) ||
                     ( (anyFarYum != null) && (plant.radius > anyFarYum.radius) )
                    )
                {
                  anyFarYum = plant;
                }
              }
            }
          } // if close/else
        } // if refuge/else
      } // if this attacks plant
      pq = pq.pickRadius(farR, viewpoint.x, viewpoint.y, plantRoot);
    } while (pq!=null);
    // end loop-through-all-within-smell-radius

    // alright, if didn't get my druthers, set best to an any
    // any's are ignored if I got what I wanted.
    if (farYumPlant==null) farYumPlant = anyFarYum;
    if ((farYumPlant==null) && (safePlant != null) &&
       ( safeD + viewpoint.radius < viewpoint.radius * viewpoint.getSmellRadius() ) )
       farYumPlant = safePlant;
    /* This version lets us choose to eat grass while hiding. Not physically right...
    if (yumPlant==null) yumPlant = anyYum;
    if (yumPlant==null) yumPlant = refugeLunch; // actually not the strategic choice....
    // if we've found anything at all in chomping range, we should have a non-null yumPlant
    if ((yumPlant == null) && (safePlant != null) &&
        (safeD < 0.0))
      yumPlant = safePlant;
    */
    // version to semi-match ScaredyGrazer of old
    // this version says if we're threatened, regardless of geometric overlap of
    // grass and refuge, we eat refuge.
    if (threat == null) {
      if (yumPlant==null) {
        if (refugeLunch != null)
          if (refugeLunch.radius >= refugeBiteR)
            yumPlant = refugeLunch; // not the strategic choice....
      }
      if (yumPlant == null) yumPlant = anyYum;
      if (yumPlant == null) yumPlant = refugeLunch;
    } else {
      if (refugeLunch != null)
        yumPlant = refugeLunch;
      if (yumPlant==null) yumPlant = anyYum;
    }

    // some more book-keeping we don't wish repeated: amHidden and distances
    // note that these distances are EDGE TO EDGE, and can often negative, meaning
    // that viewpoint's edge is INSIDE the other
    if (yumPlant != null)
      yumD = Globals.distance(yumPlant.x, yumPlant.y, viewpoint.x, viewpoint.y)
             - yumPlant.radius - viewpoint.radius;
    if (farYumPlant != null)
      farYumD = Globals.distance(farYumPlant.x, farYumPlant.y, viewpoint.x, viewpoint.y)
             - farYumPlant.radius - viewpoint.radius;
    if (safePlant != null)
      safeD = Globals.distance(safePlant.x, safePlant.y, viewpoint.x, viewpoint.y)
             - safePlant.radius - viewpoint.radius;
  }

  private void lookForSpiders() {
    Quad tq;

    Critter thing;
    Quad mobileRoot = Globals.getMobileRoot();
    double searchRadius = viewpoint.radius * viewpoint.getSmellRadius();
    double dist;
    // for safety, set null threat distance high
    threatD = Double.POSITIVE_INFINITY;

    tq = mobileRoot.pickRadius(searchRadius, viewpoint.x, viewpoint.y, mobileRoot);
    if (tq != null) do {
      thing = (Critter) tq;
      // don't use boolean attacks(me) because it says no if I'm hidden, but
      // I need to know if I need to keep hiding
      if ( thing instanceof ScaredyGrazerCarnie ) {
        threatCount++;

        // center to center distance
        dist = Globals.distance( viewpoint.x, viewpoint.y, thing.x, thing.y ) -
                thing.radius - viewpoint.radius;
        if (dist < threatD) {
          threat = thing;
          threatD = dist;
        }
      } /* if found attacker */
      tq = tq.pickRadius(searchRadius, viewpoint.x, viewpoint.y, mobileRoot);
    } while (tq!=null);
  }


}



