//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 2000, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator

/** VaryProperty is used to apply sigma specified variation
 *  to a Species progenitor parameter (or something else...using it for this
 *  for now.) Basically, VaryProperty takes a property name (like "radius"), and
 *  sigma to set it with, and a random stream, when created. Then when
 *  applied, it comes up with the next "random deviate" to order, and sets that
 *  property accordingly. VaryProperty doesn't set up the stats to observe
 *  variance. That's a separate problem. It does provide a ParamList on request
 *  to control itself via the species parameters popup menu.
 *
 *  Note that due to the case of the Species AND the individuals varying, the
 *  mean is taken as the current value (get<Propname>) of the parameter. So the
 *  deviation from the mean is found from the sigma, then added to
 *  the pre-existing mean to get the new property value. I'm getting kinda
 *  dubious about this scheme for varying progenitors, tho...
 *
 *  You can only Vary properties with double values for now. Later, may add int.
 *
 *  original, 000207 vfb, for Os, vary initial gopper body sizes in ScaredyWorld.
 */


package gecko.zoo;
import java.lang.reflect.*;
import java.beans.*;
import java.util.*;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;

public class VaryProperty {
  private String propName;
  private Class propClass;
  private Method getMethod;
  private Method setMethod;
  private double sigma;     // sqrt(variance)
  private Random randomSource;
  private boolean active = true; // defaults to active

  public VaryProperty(String propertyName, Class propertyOf, double sigma,
                      Random randomSource)
  {
    this.propName     = propertyName;
    this.propClass    = propertyOf;
    this.sigma        = sigma;
    this.randomSource = randomSource;
    findSetGetDouble();
  }

  public void setActive(boolean b){active = b;}
  public boolean isActive() {return active;}
  public void setSigma(double d) {sigma=d;}
  public double getSigma() {return sigma;}

  public boolean nameEquals(String propertyName) {
    return propName.equals(propertyName);
  }

  private void findSetGetDouble() {
    PropertyDescriptor prop;
    try {
      prop = new PropertyDescriptor(propName, propClass);
      getMethod = prop.getReadMethod();
      setMethod = prop.getWriteMethod();
    } catch (IntrospectionException e) {
      System.err.println("VaryProperty " + propName +
       ", class=" + propClass + " failed: " + e.getMessage());
    }
  }

  /** Does nothing if VaryProperty is inactive. */
  public void apply(Object toObject) {
    if (!active) return;
    try {
      Double Mean = (Double) getMethod.invoke(toObject, Param.NO_ARGS);
      double mean = Mean.doubleValue();
      double delta = randomSource.nextGaussian() * sigma;
      mean += delta;
      Object[] args = new Object[] {new Double(mean)};
      setMethod.invoke(toObject, args);
    } catch (Exception e) {
      System.err.println("VaryProperty " + propName + " failed:  " + e.getMessage());
    }
  }

  /** returns paramlist suitable for controlling this thing from a popup.
   *  If fails, returned list is empty.
   */
  public ParamList makeParam() {
    Method setActiveMethod = null;
    Method setSigmaMethod = null;
    ParamList doublet = new ParamList();
    try {
      Class[] signature = new Class[] {double.class};
      setSigmaMethod = getClass().getMethod("setSigma", signature);
      signature = new Class[] {boolean.class};
      setActiveMethod = getClass().getMethod("setActive", signature);
    } catch (Exception e) {
      System.err.println("VaryProperty " + propName + " internal error: " + e.getMessage());
      return doublet;
    }
    Param field = new Param(propName + " varies", new Boolean(active), this.getClass(),
      setActiveMethod, true);
    field.setOwner(this);
    doublet.add(field);
    field = new Param("by sigma", new Double(sigma), this.getClass(),
      setSigmaMethod, true);
    field.setOwner(this);
    doublet.add(field);

    return doublet;
  }
}